using System;
using System.Collections.Generic;
using System.Text;

namespace Avaliador.BLL
{
    public class Expressao : BusinessObject
    {
        string formula;

        public string Text
        {
            get { return formula; }
        }

        public Expressao(string expressao)
        {
            formula = expressao;
        }

        IList<Variavel> vars;

        /// <summary>
        /// permite obter a lista d evariaveis necessrias ao calculo
        /// da expresso e posteriormente atribuir um valor a cada variavel
        /// </summary>
        public IList<Variavel> Variaveis
        {
            get 
            {
                if (vars != null)
                    return vars;

                vars = new List<Variavel>();

                string[] tokens = formula.Split(' ');
                foreach (string tk in tokens)
                {
                    if (tk.StartsWith("$"))
                        vars.Add(new Variavel(tk));
                }

                return vars; 
            }
        }

        public double? valorCalculado;

        public double Valor
        {
            get 
            {
                if (valorCalculado.HasValue)
                    return valorCalculado.Value;

                IDictionary<string, double> varValues = new Dictionary<string, double>();
                foreach (Variavel v in Variaveis)
                    varValues.Add(v.Name, v.Value);

                string[] tokens = formula.Split(' ');
                Stack<string> expRPN = new Stack<string>();
                for (int i = tokens.Length - 1; i >= 0; i--)
                    expRPN.Push(tokens[i]);

                Stack<double> values = new Stack<double>();

                while (expRPN.Count > 0)
                {
                    // operador
                    if (expRPN.Peek().Length == 1 && !char.IsDigit(expRPN.Peek()[0]))
                    {
                        string op = expRPN.Pop();
                        double a1 = values.Pop();
                        double a0 = values.Pop();
                        double r = 0;
                        switch (op)
                        {
                            case "+":
                                r = a0 + a1;
                                break;
                            case "-":
                                r = a0 - a1;
                                break;
                            case "*":
                                r = a0 * a1;
                                break;
                            case "/":
                                r = a0 / a1;
                                break;
                        }
                        values.Push(r);
                    }
                    // variavel
                    else if (expRPN.Peek().StartsWith("$"))
                    {
                        values.Push(varValues[expRPN.Pop()]);
                    }
                    // constante
                    else
                    {
                        values.Push(double.Parse(expRPN.Pop()));
                    }
                }
                
                // s sobra o resultado
                valorCalculado = values.Pop();

                return valorCalculado.Value;
            }
        }

        public ResultadoCalculo CreateResultado(string designacao, string observacoes)
        {
            return new ResultadoCalculo(designacao, this, vars, observacoes);
        }
    }
}
